'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _joi = require('joi');

var _joi2 = _interopRequireDefault(_joi);

var _flattenDeep = require('lodash/flattenDeep');

var _flattenDeep2 = _interopRequireDefault(_flattenDeep);

var _pick = require('lodash/pick');

var _pick2 = _interopRequireDefault(_pick);

var _NotAllowed = require('../errors/NotAllowed');

var _NotAllowed2 = _interopRequireDefault(_NotAllowed);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

function _objectWithoutProperties(obj, keys) { var target = {}; for (var i in obj) { if (keys.indexOf(i) >= 0) continue; if (!Object.prototype.hasOwnProperty.call(obj, i)) continue; target[i] = obj[i]; } return target; }

class PermissionsManager {
  static generateCRUD(resource) {
    return ['create', 'read', 'update', 'delete'].map(action => ({
      permission: `${resource}.${action}`,
      description: `"${action}" action for the "${resource}" resource`
    }));
  }

  constructor({
    extractor = subject => subject.permissions,
    aliases = {
      view: 'read',
      access: 'read'
    }
  }) {
    this.permissions = {};
    this.aliases = aliases;
    this.extractor = extractor;
  }

  createAlias(forPermission, toPermission) {
    this.aliases[forPermission] = toPermission;
  }

  getAll() {
    return Object.keys(this.permissions).reduce((acc, permission) => [...acc, _extends({
      permission
    }, (0, _pick2.default)(this.permissions[permission], ['description', 'dependencies']))], []);
  }

  has(name) {
    return Object.keys(this.permissions).includes(name);
  }

  define(permission, rawConfig = {}) {
    if (Array.isArray(permission)) {
      return permission.forEach((_ref) => {
        let { permission: permissionName } = _ref,
            itemConfig = _objectWithoutProperties(_ref, ['permission']);

        return this.define(permissionName, _extends({}, rawConfig, itemConfig));
      });
    }

    const config = _joi2.default.attempt(rawConfig, {
      description: _joi2.default.string().default(`${permission} permission definition`),
      dependencies: _joi2.default.array().default([]),
      check: _joi2.default.func().default(() => true),
      isValidObject: _joi2.default.func().default(() => true)
    });

    if (this.has(permission)) {
      throw new Error(`Can't add permission with a name of ${permission}. Already exists!`);
    }

    const dependencies = (0, _flattenDeep2.default)(config.dependencies);

    dependencies.forEach(dependency => {
      if (!this.has(dependency)) {
        throw new Error(`Unable to find permission named "${dependency}" required as a dependency of "${permission}".`);
      }
    });

    this.permissions[permission] = config;

    return this;
  }

  can(subject, permissionName, object) {
    var _this = this;

    return _asyncToGenerator(function* () {
      try {
        yield _this.check(subject, permissionName, object);
      } catch (err) {
        if (err instanceof _NotAllowed2.default) {
          return false;
        }

        throw err;
      }

      return true;
    })();
  }

  checkAll(subject, permissions, object, parallel = true) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      const walkPermissions = permissions.map(function (item) {
        if (Array.isArray(item)) {
          return function () {
            return _this2.checkAll(subject, item, object, !parallel);
          };
        }
        return function () {
          return _this2.check(subject, item, object);
        };
      });

      return parallel ? Promise.all(walkPermissions.map(function (task) {
        return task();
      })) : walkPermissions.reduce(function (promise, task) {
        return promise.then(task);
      }, Promise.resolve());
    })();
  }

  check(subject, permissionName, object) {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      (0, _assert2.default)(subject, 'Subject is required!');
      (0, _assert2.default)(permissionName && typeof permissionName === 'string', 'Permission is required!');

      const name = Object.keys(_this3.aliases).includes(permissionName) ? _this3.aliases[permissionName] : permissionName;

      if (!_this3.has(name)) {
        throw new Error(`Unable to find permission named "${name}" when checking access.`);
      }

      const permission = _this3.permissions[name];

      const subjectPermissions = yield _this3.extractor(subject);

      if (!subjectPermissions.includes(name)) {
        throw new _NotAllowed2.default(`Failed permission check "${name}"! Reason: permission not included in the list of subjects permissions (${subjectPermissions}).`);
      }

      if (!permission.isValidObject(object)) {
        throw new _NotAllowed2.default(`Failed permission check "${name}"! Reason: received invalid object (${object}).`);
      }

      yield _this3.checkAll(subject, permission.dependencies, object);

      const isAllowed = yield permission.check(subject, object);

      if (!isAllowed) {
        throw new _NotAllowed2.default(`Failed permission check "${name}"! Reason: failed check function.`);
      }
    })();
  }
}

exports.default = PermissionsManager;